#include "ply.h"

struct PlyVertex
{
	float x, y, z;
	//unsigned char c;
	float c;
	unsigned char r, g, b;
	unsigned char m;
};

struct PlyFace
{
	unsigned char nverts;
	int *verts;
};

template <class floating>
bool Mesh<floating>::readPly(const char *filename, MeshMode *meshMode)
{	
	PlyFile *ply = NULL;
	int nelems;
	char **elemNames = NULL;	

	FILE *fp = fopen(filename, "rb");
	if (fp)
	{
		ply = ply_read(fp, &nelems, &elemNames);
		if (ply == NULL) 
		{
			cerr << "Mesh::readPly: fail to read " << filename << endl;
			return false;
		}
		meshMode->_ascii = (ply->file_type == PLY_ASCII);
	}
	else
	{
		cerr << "Mesh::readPly: fail to open " << filename << endl;
		return false;
	}	

	// go through each kind of element that we learned is in the file 
	// and read them
	char *elemName;
	PlyProperty **plist;
	int numElems, nprops;
	int i, j;
	for (i = 0; i < nelems; i++) 
	{
		// get the description of the first element
		elemName = elemNames[i];
		plist = ply_get_element_description(ply, elemName, &numElems, &nprops);

		int xyz = 0;
		int conf = 0;		
		int rgb = 0;		
		int mask = 0;
		int vind = 0;
		for (j = 0; j < nprops; j++)
		{		
			plist[j]->internal_type = plist[j]->external_type;

			if (strcmp(plist[j]->name, "x") == 0) 
			{
				plist[j]->offset = offsetof(PlyVertex, x);
				xyz ++;
			}
			else if (strcmp(plist[j]->name, "y") == 0) 
			{
				plist[j]->offset = offsetof(PlyVertex, y);
				xyz ++;
			}
			else if (strcmp(plist[j]->name, "z") == 0)
			{
				plist[j]->offset = offsetof(PlyVertex, z);
				xyz ++;
			}
			else if (strcmp(plist[j]->name, "confidence") == 0)
			{				
				plist[j]->offset = offsetof(PlyVertex, c);
				conf ++;
			}
			else if (strcmp(plist[j]->name, "diffuse_red") == 0) 
			{ 
				plist[j]->offset = offsetof(PlyVertex, r);
				rgb ++;
			}
			else if (strcmp(plist[j]->name, "diffuse_green") == 0) 
			{
				plist[j]->offset = offsetof(PlyVertex, g);
				rgb ++;
			}
			else if (strcmp(plist[j]->name, "diffuse_blue") == 0)
			{
				plist[j]->offset = offsetof(PlyVertex, b);
				rgb ++;
			}
			else if (strcmp(plist[j]->name, "mask") == 0)
			{				
				plist[j]->offset = offsetof(PlyVertex, m);
				mask ++;
			}

			if (strcmp(plist[j]->name, "vertex_indices") == 0)
			{				
				plist[j]->offset = offsetof(PlyFace, verts);
				plist[j]->count_external = plist[j]->count_internal = PLY_UCHAR;
				plist[j]->count_offset = offsetof(PlyFace, nverts);
				vind ++;
			}
			else 
			{
				plist[j]->count_external = plist[j]->count_internal = 0;
				plist[j]->count_offset = 0;
			}			

			if (ply_get_property(ply, elemName, plist[j]) == 0)
			{
				cerr << "Mesh::readPly: fail to read " << plist[j]->name << " in " << filename << endl;
				return false;
			}

			free(plist[j]->name);				
			free(plist[j]);
		}
		free(plist);		

		// if we're on vertex elements, read them in
		if (strcmp("vertex", elemName) == 0) 
		{
			if (xyz != 3)
			{
				cerr << "Mesh::readPly: fail to read xyz in " << filename << endl;
				return false;
			}

			if (meshMode->_hasConf == -1)
			{
				meshMode->_hasConf = (conf == 1);
			}
			else if (meshMode->_hasConf == 1 &&	conf != 1)
			{
				cerr << "Mesh::readPly: fail to read confidence in " << filename << endl;
				return false;
			}

			if (meshMode->_hasColor == -1)
			{
				meshMode->_hasColor = (rgb == 3);
			}
			else if (meshMode->_hasColor ==1 && rgb != 3)
			{
				cerr << "Mesh::readPly: fail to read rgb in " << filename << endl;
				return false;
			}

			if (meshMode->_hasMask == -1)
			{
				meshMode->_hasMask = (mask == 1);
			}
			else if (meshMode->_hasMask == 1 && mask != 1)
			{
				cerr << "Mesh::readPly: fail to read mask in " << filename << endl;
				return false;
			}

			_verts.resize(numElems);
			// grab all the vertex elements
			for (j = 0; j < numElems; j++) 
			{
				// grab an element from the file
				PlyVertex vert;
				ply_get_element (ply, &vert);
				Vertex <floating> &mv = _verts[j];
				mv._coord[0] = (floating)vert.x;
				mv._coord[1] = (floating)vert.y;
				mv._coord[2] = (floating)vert.z;
				if (meshMode->_hasConf)
				{
					mv._confidence = vert.c;
				}
				if (meshMode->_hasColor)
				{
					mv._matDiff[0] = vert.r;
					mv._matDiff[1] = vert.g;
					mv._matDiff[2] = vert.b;
				}
				if (meshMode->_hasMask)
				{
					mv._mask = vert.m;
				}
			}
		}		
		// if we're on face elements, read them in
		else if (strcmp("face", elemName) == 0) 
		{
			if (vind != 1)
			{
				cerr << "Mesh::readPly: fail to read vertex_indices in " << filename << endl;
				return false;
			}
			if (meshMode->_hasTris == 0)
			{
				continue;
			}
			else if (meshMode->_hasTris == -1)
			{
				meshMode->_hasTris = 1;
			}

			_tris.resize(numElems);
			// grab all the face elements
			for (j = 0; j < numElems; j++) 
			{
				// grab an element from the file
				PlyFace face;
				ply_get_element(ply, &face);

				Triangle <floating> &mt = _tris[j];
				mt._vindex1 = face.verts[0];
				mt._vindex2 = face.verts[1];
				mt._vindex3 = face.verts[2];

				free(face.verts);
			}
		}

		free(elemName);
	}
	free(elemNames);

	ply_close(ply);

	_name = filename;
	return true;
}

template <class floating>
bool Mesh<floating>::writePly(const char *filename, const MeshMode *meshMode) const
{
	PlyFile *ply;	
	FILE *fp = fopen(filename, "wb");
	if (fp)
	{
		char *elem_names[] = { // list of the kinds of elements in the user's object 
			"vertex", "face"
		};
		ply = ply_write (fp, 
			meshMode->_hasTris ? 2 : 1, 
			elem_names, 
			meshMode->_ascii ? PLY_ASCII : PLY_BINARY_LE);
	}
	else
	{
		cerr << "Mesh::writePly: fail to open " << filename << endl;
		return false;
	}

	PlyProperty vert_props[8] = { // list of property information for a vertex 				
		{"x", PLY_FLOAT, PLY_FLOAT, offsetof(PlyVertex,x), 0, 0, 0, 0},
		{"y", PLY_FLOAT, PLY_FLOAT, offsetof(PlyVertex,y), 0, 0, 0, 0},
		{"z", PLY_FLOAT, PLY_FLOAT, offsetof(PlyVertex,z), 0, 0, 0, 0},
		//whether to write the following prop depends on meshMode
		//{"confidence", PLY_FLOAT, PLY_FLOAT, offsetof(PlyVertex,c), 0, 0, 0, 0},
		//{"diffuse_red", PLY_UCHAR, PLY_UCHAR, offsetof(PlyVertex,r), 0, 0, 0, 0},
		//{"diffuse_green", PLY_UCHAR, PLY_UCHAR, offsetof(PlyVertex,g), 0, 0, 0, 0},
		//{"diffuse_blue", PLY_UCHAR, PLY_UCHAR, offsetof(PlyVertex,b), 0, 0, 0, 0},
		//{"mask", PLY_UCHAR, PLY_UCHAR, offsetof(PlyVertex,m), 0, 0, 0, 0},
	};
	int nprops = 3;

	// describe what properties go into the vertex elements 
	if (meshMode->_hasConf)
	{
		PlyProperty *p = vert_props + nprops;
		p->name = "confidence";
		p->external_type = p->internal_type = PLY_FLOAT;
		p->offset = offsetof(PlyVertex,c);
		p->is_list = p->count_external = p->count_internal = p->count_offset = 0;
		nprops ++;
	}
	if (meshMode->_hasColor)
	{
		PlyProperty *p;
		
		p = vert_props + nprops;
		p->name = "diffuse_red";
		p->external_type = p->internal_type = PLY_UCHAR;
		p->offset = offsetof(PlyVertex,r);
		p->is_list = p->count_external = p->count_internal = p->count_offset = 0;
		nprops ++;

		p = vert_props + nprops;
		p->name = "diffuse_green";
		p->external_type = p->internal_type = PLY_UCHAR;
		p->offset = offsetof(PlyVertex,g);
		p->is_list = p->count_external = p->count_internal = p->count_offset = 0;
		nprops ++;

		p = vert_props + nprops;
		p->name = "diffuse_blue";
		p->external_type = p->internal_type = PLY_UCHAR;
		p->offset = offsetof(PlyVertex,b);
		p->is_list = p->count_external = p->count_internal = p->count_offset = 0;
		nprops ++;
	}
	if (meshMode->_hasMask)
	{
		PlyProperty *p = vert_props + nprops;
		p->name = "mask";
		p->external_type = p->internal_type = PLY_UCHAR;
		p->offset = offsetof(PlyVertex,m);
		p->is_list = p->count_external = p->count_internal = p->count_offset = 0;
		nprops ++;
	}
    
	ply_describe_element(ply, "vertex", (int)_verts.size(), nprops, vert_props);

	if (meshMode->_hasTris)
	{
		PlyProperty face_props[] = { // list of property information for a face 
			{"vertex_indices", PLY_INT, PLY_INT, offsetof(PlyFace,verts),
				1, PLY_UCHAR, PLY_UCHAR, offsetof(PlyFace,nverts)},
		};

		ply_describe_element(ply, "face", (int)_tris.size(), 1, face_props);
	}
    
	ply_header_complete(ply);

	int i;

	// set up and write the vertex elements 
	ply_put_element_setup(ply, "vertex");
	PlyVertex pv;
	for (i = 0; i < (int)_verts.size(); i++)
	{
		const Vertex <floating> &mv = _verts[i];
		
		pv.x = (float)mv._coord[0];
		pv.y = (float)mv._coord[1];
		pv.z = (float)mv._coord[2];
		if (meshMode->_hasConf)
		{
			pv.c = mv._confidence;
		}
		if (meshMode->_hasColor)
		{
			pv.r = mv._matDiff[0];
			pv.g = mv._matDiff[1];
			pv.b = mv._matDiff[2];
		}
		if (meshMode->_hasMask)
		{
			pv.m = mv._mask;
		}
		ply_put_element(ply, (void *)&pv);
	}

	if (meshMode->_hasTris)
	{
		// set up and write the face elements 
		ply_put_element_setup (ply, "face");		
		PlyFace pf;
		int vindex[3];
		pf.nverts = 3;
		pf.verts = vindex;
		for (i = 0; i < (int)_tris.size(); i++)
		{
			const Triangle <floating> &mt = _tris[i];
			pf.verts[0] = mt._vindex1;
			pf.verts[1] = mt._vindex2;
			pf.verts[2] = mt._vindex3;
			ply_put_element(ply, (void *) &pf);
		}
	}

	//close file, free memory pointed by ply
	ply_close(ply);

	return true;
}

